{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "8b1cad59-f261-4437-865c-044faa88d611",
   "metadata": {},
   "source": [
    "# Advanced feature selection with R\n",
    "\n",
    "This notebook shows how you can use R to accomplish feature selection by creating aggregated feature impact. A [Python notebook](https://github.com/datarobot-community/examples-for-data-scientists/blob/master/Feature%20Lists%20Manipulation/Python/Advanced%20Feature%20Selection.ipynb) is also available.\n",
    "\n",
    "## Requirements\n",
    "\n",
    "To use the code provided in this tutorial, make sure you have the following:\n",
    "\n",
    "* R version 3.6.2+.\n",
    "* [The DataRobot R package](https://cran.r-project.org/package=datarobot).\n",
    "* DataRobot API version 2.2.0+.\n",
    "* A DataRobot `Project` object in R.\n",
    "* A DataRobot `Model` object in R."
   ]
  },
  {
   "cell_type": "markdown",
   "id": "79d41538-d975-425b-8f42-0ca6789aa690",
   "metadata": {},
   "source": [
    "### Install packages and connect to DataRobot\n",
    "\n",
    "Run the following command to install the required libraries and connect to DataRobot by providing your endpoint and [API key](https://docs.datarobot.com/en/docs/platform/account-mgmt/acct-settings/api-key-mgmt.html)."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "98f11e6d-880b-4b61-8588-8273f673f40b",
   "metadata": {},
   "outputs": [],
   "source": [
    "library(datarobot)\n",
    "library(dplyr)\n",
    "library(stringr)\n",
    "library(ggplot2)\n",
    "library(purrr)\n",
    "\n",
    "ConnectToDataRobot(endpoint = \"YOUR ENDPOINT\",\n",
    "                   token = \"YOUR TOKEN\")"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "c9304065-e68f-45cc-a26c-7ca68065c939",
   "metadata": {},
   "source": [
    "### List models\n",
    "\n",
    "After installing the libraries above, select the project you want to use and get its project ID. Then, use the code below to list all of the models built for that project."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "d92141f3-2260-4c45-af10-f8f82e683393",
   "metadata": {},
   "outputs": [],
   "source": [
    "project <- GetProject(\"YOUR_PROJECT_ID\")\n",
    "allModels <- ListModels(project)\n",
    "modelFrame <- as.data.frame(allModels)\n",
    "metric <- modelFrame$validationMetric\n",
    "if (project$metric %in% c('AUC', 'Gini Norm')) {\n",
    "  bestIndex <- which.max(metric)\n",
    "} else {\n",
    "  bestIndex <- which.min(metric)\n",
    "}\n",
    "model <- allModels[[bestIndex]]\n",
    "model$modelType"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "9af8f388-fd0b-4132-8725-5c983905bfc1",
   "metadata": {},
   "source": [
    "### Filter the model list\n",
    "\n",
    "Next, choose which models to use for the Feature Impact calculation. This tutorial only uses highly ranked, non-blender or auto-tuned models to optimize for speed and computational power. However, you can include these types of models if runtime is not a concern for your project. DataRobot recommends that you not select models trained on small sample sizes; instead, use models trained on 64% and 80% of the data.\n",
    "\n",
    "The following code filters the model list to only use the top-performing models, trained on over 63% of the data. However, you can take alternative approaches, such as only using models within one modeling family or increasing the sample size."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "fdfdca9a-d1cc-4c45-bdca-5600d5e0d7ae",
   "metadata": {},
   "outputs": [],
   "source": [
    "models <- ListModels(project)\n",
    "bestmodels <- Filter(function(m) m$featurelistName == \"Informative Features\" & m$samplePct >= 64 &  m$samplePct <= 80 & !str_detect(m$modelType, 'Blender') & !str_detect(m$modelType, 'Auto-Tuned') , models)\n",
    "bestmodels <- Filter(function(m) m$samplePct >= 63, models)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "2316be4c-cf82-41a9-91ae-c56990976ea7",
   "metadata": {},
   "source": [
    "### Get Feature Impact\n",
    "\n",
    "Now, calculate Feature Impact for selected models. This `for` loop iterates over the list and isolates the feature impact data."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "f7fa913c-a3bc-43a7-9ff7-8203005ce58d",
   "metadata": {},
   "outputs": [],
   "source": [
    "all_impact<- NULL\n",
    "for(i in 1:length(bestmodels)) {  \n",
    "    featureImpact <- GetFeatureImpact(bestmodels[[i]])\n",
    "    featureImpact$modelname <- bestmodels[[i]]$modelType\n",
    "    all_impact <- rbind(all_impact,featureImpact)\n",
    "  }"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "51067639-01de-4db8-9444-143a1903e8ea",
   "metadata": {},
   "source": [
    "### Plot Feature Impact\n",
    "\n",
    "You can plot the project's features and their medians using GGplot’s `geom_boxplot` (displayed below). This tutorial plots the boxplots for unnormalized feature impact; however, there are other ways to display this information, such as plotting the standard deviation."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "7b732375-5f3a-4eef-be85-acdd4f6ff451",
   "metadata": {},
   "outputs": [],
   "source": [
    "all_impact <- all_impact %>% mutate(finalnorm = impactUnnormalized/max(impactUnnormalized))\n",
    "p <- ggplot(all_impact, aes(x=reorder(featureName, finalnorm, FUN=median), y=finalnorm))\n",
    "p + geom_boxplot(fill= \"#2D8FE2\") + coord_flip() + theme(axis.text=element_text(size=16),\n",
    "        axis.title=element_text(size=12,face=\"bold\"))"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "97cf70c6-e6c8-4fcb-aede-7bf7db7a4859",
   "metadata": {},
   "source": [
    "![](r-fs-1.png)\n",
    "\n",
    "This plot helps you identify the features that are driving model decisions the most.\n",
    "\n",
    "### Create a new feature list\n",
    "\n",
    "After computing the most impactful features, create a new feature list that includes only those features. You can check the application to see the feature list in the project."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "10ec983f-1e82-425f-8c89-41dea522064c",
   "metadata": {},
   "outputs": [],
   "source": [
    "ranked_impact <- all_impact %>% group_by(featureName) %>%\n",
    "    summarise(impact = mean(finalnorm)) %>%\n",
    "    arrange(desc(impact))\n",
    "\n",
    "topfeatures <- pull(ranked_impact,featureName)\n",
    "No_of_features_to_select <- 10\n",
    "listname = paste0(\"TopFI_\", No_of_features_to_select)\n",
    "Feature_id_percent_rank = CreateFeaturelist(project, listName= listname , featureNames = topfeatures[1:No_of_features_to_select])$featurelistId"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "a574864f-3e9f-4a51-b8e8-1fe66f6a1bb0",
   "metadata": {},
   "source": [
    "### Re-run Autopilot\n",
    "\n",
    "Now, you can rerun Autopilot using the new feature list to see how performance changes with the application of the new list.\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "ea657ffa-6e89-481d-baf4-b8442a99b9f7",
   "metadata": {},
   "outputs": [],
   "source": [
    "StartNewAutoPilot(project,featurelistId = Feature_id_percent_rank)\n",
    "WaitForAutopilot(project)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "d6f3d71a-3c84-44f1-a776-b7773b754773",
   "metadata": {},
   "source": [
    "You can create feature lists of varying lengths, plot the performance for each, and compare. This can help you decide how many features to include when building your model.\n",
    "\n",
    "The code below creates a loop that runs Autopilot using the top 9, 6, 3, and 1 feature(s)."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "69bd3b6d-dda5-45d7-ad46-b548f53a407e",
   "metadata": {},
   "outputs": [],
   "source": [
    "No_of_features_to_loop <- c(9, 6, 3, 1)\n",
    "\n",
    "for(i in 1:length(No_of_features_to_loop)) {\n",
    "  listname = paste0(\"TopFI_\", No_of_features_to_loop[i])\n",
    "  Feature_id_percent_rank = CreateFeaturelist(project, listName= listname , featureNames = topfeatures[1:i])$featurelistId\n",
    "  StartNewAutoPilot(project,featurelistId = Feature_id_percent_rank)\n",
    "  WaitForAutopilot(project)\n",
    "}"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "a48bede5-a512-4f48-8841-6ebbe76bd061",
   "metadata": {},
   "source": [
    "### Measure and plot AUC scores\n",
    "\n",
    "As a final step, measure the Area Under the Curve (AUC) score for each model trained on the new feature list. AUC is a common error metric for binary classification projects that considers all possible thresholds and summarizes performance in a single value on the ROC Curve."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "9f475267-d7e0-4ebc-996c-02b4c8271121",
   "metadata": {},
   "outputs": [],
   "source": [
    "{r eval=FALSE}\n",
    "#Create a list of models and select the ones that used the new feature lists.\n",
    "models <- ListModels(project)\n",
    "FI_list <- c(\"TOPFI_1\", \"TOPFI_3\", \"TOPFI_6\", \"TOPFI_9\")\n",
    "unpack_func <- function(i) {\n",
    "  unpack <- models[[i]]\n",
    "  modelid <- unpack$modelId\n",
    "  auc <- unpack$metrics$AUC$validation\n",
    "  fname <- unpack$featurelistName\n",
    " output <- c(modelid, auc, fname)\n",
    "  return(output)\n",
    "}\n",
    "# map over function and convert to dataframe\n",
    "fi_models <- map(1:length(models), unpack_func)\n",
    "output <- as.data.frame(t(sapply(fi_models, function(x) x[1:3])))\n",
    "output$V2 <- as.numeric(as.character(output$V2))\n",
    "#find the max score for each feature list.\n",
    "result <- output %>%\n",
    "             group_by(V3) %>%\n",
    "             filter(V2 == max(V2)) %>%\n",
    "            filter(!duplicated(V3))\n",
    "#Filter only the ones we want to plot\n",
    "graph <-  filter(result, V3 == \"TopFI_1\" |  V3 == \"TopFI_3\"  | V3 == \"TopFI_6\" |  V3 == \"TopFI_9\")\n",
    "#Create numeric for the number of features in each list.\n",
    "graph$number_of_features <- c(1, 3, 6, 9)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "872b0d2a-6d17-46ef-adb2-a03128f41b8b",
   "metadata": {},
   "source": [
    "The code below is a custom function that pulls this data from an updated model list and then graphs it using `ggplot` (displayed below)."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "8472c4ab-3aa1-473b-bad5-1ba5a2c6a005",
   "metadata": {},
   "outputs": [],
   "source": [
    "{r eval=FALSE}\n",
    "ggplot(data=graph, aes(x=graph$number_of_features, y=graph$V2) ) +\n",
    "  geom_line(color = \"#2D8FE2\")+\n",
    "  geom_point(color = \"#2D8FE2\" )+\n",
    "  theme(axis.text=element_text(size=16), axis.title=element_text(size=12,face=\"bold\")) +\n",
    "  labs(title= \"AUC Validation Scores\") +\n",
    "  xlab(\"Number of Features\") +\n",
    "  ylab(\"AUC\") +\n",
    "  ylim(0, 0.8)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "b4a0dfc5-9cbf-4999-a1be-b1f38403bbe4",
   "metadata": {},
   "source": [
    "![](r-fs-2.png)\n",
    "\n",
    "You can see in this example that increasing the size of the feature list lowers the model's AUC score.\n",
    "\n",
    "## Documentation\n",
    "\n",
    "* [Feature Impact](https://docs.datarobot.com/en/docs/modeling/analyze-models/understand/feature-impact.html)\n",
    "* [Feature Lists](https://docs.datarobot.com/en/docs/modeling/build-models/build-basic/feature-lists.html)\n",
    "* [AUC score](https://docs.datarobot.com/en/docs/glossary/index.html#auc-area-under-the-curve)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "3f5ab481-3fec-42f2-ba61-7e159e324163",
   "metadata": {},
   "outputs": [],
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "R",
   "language": "R",
   "name": "ir"
  },
  "language_info": {
   "codemirror_mode": "r",
   "file_extension": ".r",
   "mimetype": "text/x-r-source",
   "name": "R",
   "pygments_lexer": "r",
   "version": "4.1.2"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}
